import os
import shutil
import unittest
from pprint import pprint

from ds_core.handlers.abstract_handlers import ConnectorContract, HandlerFactory
from ds_core.properties.property_manager import PropertyManager


class PropertyManagerTest(unittest.TestCase):
    """Test: """

    def setUp(self):
        try:
            shutil.rmtree('test')
        except:
            pass
        os.mkdir('test')
        pass

    def tearDown(self):
        PropertyManager()._remove_all()
        self.assertEqual({}, PropertyManager().get_all())

        try:
            shutil.rmtree('test')
        except:
            pass

    def test_runs(self):
        """Basic smoke test"""
        PropertyManager()

    def test_replace_dict_with_dict(self):
        pm = PropertyManager()
        control = {'contract': {'connector': {}, 'cleaner': '{}', 'version': '0.00'}}
        result = {'root': {'contract': {'connector': {}, 'cleaner': '{}', 'version': '0.00'}}}
        pm.set('root', control)
        self.assertEqual(result, pm.get_all())
        control = {'contract': {'connector': {'files': ['myfile.csv', 'yourfile.csv']}, 'version': '1.00.001'}}
        result = {'root': {'contract': {'connector': {'files': ['myfile.csv', 'yourfile.csv']},
                                        'cleaner': '{}',
                                        'version': '1.00.001'}}}
        pm.set('root', control)
        self.assertEqual(result, pm.get_all())
        control = {'contract': {'connector': {'files': {'in' : 'in_file', 'out': 'out_file'}},
                                'version': '1.00.002'}}
        result = {'root': {'contract': {'connector': {'files': {'in' : 'in_file', 'out': 'out_file'}},
                                        'cleaner': '{}',
                                        'version': '1.00.002'}}}
        pm.set('root', control)
        self.assertEqual(result, pm.get_all())

    def test_set_list(self):
        pm = PropertyManager()
        pm.set('some_name', ['item1'])
        self.assertEqual(['item1'], pm.get('some_name'))
        pm.set('some_name', ['item2', 'item3'])
        self.assertEqual(['item1', 'item2', 'item3'], pm.get('some_name'))

    def test_dump_load_persist(self):
        connector_contract = ConnectorContract(uri='event://pm_story',
                                               module_name='ds_core.handlers.event_handlers',
                                               handler='EventPersistHandler')
        handler = HandlerFactory.instantiate(connector_contract)
        pm = PropertyManager()
        file = 'test/properties.parquet'
        pm.set('KeyA', 'ValueA')
        self.assertEqual('ValueA', pm.get('KeyA'))
        pm.dump(handler)
        self.assertTrue(pm.is_key('KeyA'))
        pm.remove('KeyA')
        self.assertFalse(pm.is_key('KeyA'))
        pm.load(handler)
        self.assertTrue(pm.is_key('KeyA'))
        self.assertEqual('ValueA', pm.get('KeyA'))

    def test_dump_load_key(self):
        connector_contract = ConnectorContract(uri='event://pm_story',
                                               module_name='ds_core.handlers.event_handlers',
                                               handler='EventPersistHandler')
        handler = HandlerFactory.instantiate(connector_contract)
        pm = PropertyManager()
        pm.set('transition.task1', {'keyA1': 'valueA1'})
        pm.set('transition.task2', {'keyA2': 'valueA2'})
        pm.set('feature_catalog.task2', {'keyA2': 'valueA2'})
        base = pm.get_all()
        pm.dump(handler=handler)
        pm.load(handler=handler, replace=True)
        self.assertDictEqual(base, pm.get_all())
        pm.load(handler=handler, key='transition.task1')
        control = {'transition': {'task1': {'keyA1': 'valueA1'}, 'task2': {'keyA2': 'valueA2'}}, 'feature_catalog': {'task2': {'keyA2': 'valueA2'}}}
        self.assertDictEqual(control, pm.get_all())
        pm.load(handler=handler, key='transition')
        control = {'transition': {'task1': {'keyA1': 'valueA1'}, 'task2': {'keyA2': 'valueA2'}}, 'feature_catalog': {'task2': {'keyA2': 'valueA2'}}}
        self.assertDictEqual(control, pm.get_all())
        pm.load(handler=handler, key='', ignore_key_error=True)
        self.assertDictEqual(control, pm.get_all())
        pm.load(handler=handler, key='one.two.three', ignore_key_error=True)
        self.assertDictEqual(control, pm.get_all())

    def test_raise(self):
        connector_contract = ConnectorContract(uri='event://pm_story',
                                               module_name='ds_core.handlers.event_handlers',
                                               handler='EventPersistHandler')
        handler = HandlerFactory.instantiate(connector_contract)
        pm = PropertyManager()
        pm.dump(handler)
        with self.assertRaises(KeyError) as context:
            pm.load(handler=handler, key='one.two.three')
        self.assertTrue("The key 'one.two.three' could not " in str(context.exception))

    def test_dump_load_overwrite(self):
        connector_contract = ConnectorContract(uri='event://pm_story',
                                               module_name='ds_core.handlers.event_handlers',
                                               handler='EventPersistHandler')
        handler = HandlerFactory.instantiate(connector_contract)
        pm = PropertyManager()
        pm.set('KeyA', 'ValueA')
        self.assertEqual('ValueA', pm.get('KeyA'))
        # dump the file then add a new item and see if it doesn't get overwritten
        pm.dump(handler)
        pm.set('KeyB', 'ValueB')
        self.assertEqual('ValueB', pm.get('KeyB'))
        pm.remove('KeyA')
        self.assertFalse(pm.is_key('KeyA'))
        self.assertTrue(pm.is_key('KeyB'))
        pm.load(handler)
        self.assertTrue(pm.is_key('KeyA'))
        self.assertEqual('ValueA', pm.get('KeyA'))
        self.assertTrue(pm.is_key('KeyB'))
        self.assertEqual('ValueB', pm.get('KeyB'))
        # change KeyA to another value and load again
        pm.set('KeyA', 'ReplaceA')
        self.assertEqual('ReplaceA', pm.get('KeyA'))
        pm.load(handler)
        self.assertEqual('ValueA', pm.get('KeyA'))
        self.assertEqual('ValueB', pm.get('KeyB'))

    def test_singleton(self):
        """tests singleton works"""
        pm1 = PropertyManager()
        pm2 = PropertyManager()

        # adding properties to s1 exists in s2
        pm1.set('KeyA', 'ValueA')
        pm1.set('KeyB', 'ValueB')
        self.assertEqual(pm1.get_all(), pm2.get_all())
        self.assertEqual(pm1.get('KeyA'), 'ValueA')
        self.assertEqual(pm2.get('KeyA'), 'ValueA')
        self.assertEqual(pm1.get('KeyB'), 'ValueB')
        self.assertEqual(pm2.get('KeyB'), 'ValueB')

        # removing properties
        pm1.remove('KeyB')
        self.assertEqual(pm1.get('KeyB'), None)
        self.assertEqual(pm2.get('KeyB'), None)

        # add property to s2
        pm2.set('KeyC', 'ValueC')
        self.assertEqual(pm1.get('KeyC'), 'ValueC')

    def test_is_key(self):
        """tests singleton works"""
        pm = PropertyManager()
        pm.set('level0.level1.level2', 'value012')

        self.assertFalse(pm.is_key('NoKey'))
        self.assertFalse(pm.is_key(''))
        self.assertFalse(pm.is_key(None))

        self.assertTrue(pm.is_key('level0'))
        self.assertTrue(pm.is_key('level0.level1'))
        self.assertTrue(pm.is_key('level0.level1.level2'))
        self.assertFalse(pm.is_key('level0.level1.notKey'))

    def test_get(self):
        """ Test the get with complex dictionary"""
        pm = PropertyManager()
        pm.set('level0.level1.branch1', '01b1')
        pm.set('level0.level1.branch2', '01b2')

        self.assertEqual(pm.get(''), None)
        self.assertEqual(pm.get('noValue'), None)
        self.assertEqual(pm.get('noValue.noValue'), None)
        self.assertEqual(pm.get('level0.noValue'), None)
        self.assertEqual(pm.get('level0.level1.noValue'), None)

        self.assertEqual('01b1', pm.get('level0.level1.branch1'))
        self.assertEqual('01b2', pm.get('level0.level1.branch2'))
        self.assertEqual({'branch1': '01b1', 'branch2': '01b2'}, pm.get('level0.level1'))

    def test_replace_list(self):
        # Configuration
        pm = PropertyManager()
        pm.set('level0.level1', {'keyList': ['list1', 'list2', 'list3'], 'newItem': 'newValue'})
        self.assertEqual(['list1', 'list2', 'list3'], pm.get('level0.level1.keyList'))
        self.assertEqual('newValue', pm.get('level0.level1.newItem'))
        pm.set('level0.level1', {'keyList': ['list5', 'list6', 'list7'], 'newItem': 'newValue'})
        self.assertNotEqual(['list1', 'list2', 'list3'], pm.get('level0.level1.keyList'))
        self.assertEqual(['list5', 'list6', 'list7'], pm.get('level0.level1.keyList'))

    def test_join(self):
        pm = PropertyManager()
        self.assertEqual('1.2.3', pm.join('1', '2', '3'))
        self.assertEqual('1/2/3', pm.join('1', '2', '3', sep='/'))


if __name__ == '__main__':
    unittest.main()



import ast
import threading
import uuid
from datetime import datetime
from typing import Any
import pyarrow as pa
import ds_core
from ds_core.handlers.abstract_handlers import AbstractPersistHandler
from ds_core.properties.decorator_patterns import singleton

class PropertyManager(object):
    """

    A thread safe singleton in-memory, NoSQL, multi-tenant class that allows for the management of key/value pairs
    to be stored and retrieved. The persisted values are stored in files specified when persisted

    The value's are stored as a tree structure with the key being a dot separated string value up the tree.
    For example key 'root.directories.base_dir' returns [str]: 'filepath' .
    Where the underlying Dictionary looks like { root: { directories: { base_dir : 'filepath' }}}.

    The key can reference any part of the tree and will return the object at that point.
    From the example above key 'root' would return [dict]: { directories: { base_dir : 'filepath' }}.

    The key must start from the base key and work up, this allows namespace to avoid repeated key values.
    """

    __properties = dict({})
    __lock = threading.Lock()

    @singleton
    def __new__(cls):
        return super().__new__(cls)

    @classmethod
    def is_key(cls, key: Any) -> bool:
        """identifies if a key exists or not.

        :param key: the key of the value
            The key should be a dot separated string of keys from root up the tree
        :return:
            True if the key exists in the properties
            False if the key doesn't exist in the properties
        """
        if key is None or not key:
            return False
        find_dict = cls.__properties
        is_path, _, is_key = key.rpartition('.')
        if is_path:
            for part in is_path.split('.'):
                if isinstance(find_dict, dict):
                    find_dict = find_dict.get(part, {})
                else:
                    break
        if is_key in find_dict:
            return True
        return False

    @classmethod
    def get(cls, key: Any, default: Any=None) -> [object, str, dict, tuple, list, int, float]:
        """ gets a property value for the dot separated key.

        :param key: the key of the value
            The key should be a dot separated string of keys from root up the tree
        :param default: (optional) a default value to return if no value is found

        :return:
            an object found in the key can be any structure found under that key
            if the key is not found, None is returned
            If the key is None then the complete properties dictionary is returned
            will be the full tree under the requested key, be it a value, tuple, list or dictionary
        """
        if key is None or not key:
            return default
        rtn_val = cls.__properties
        for part in key.split('.'):
            if isinstance(rtn_val, dict):
                rtn_val = rtn_val.get(part)
                if rtn_val is None:
                    return default
            else:
                return default
        with cls.__lock:
            return rtn_val

    @classmethod
    def get_all(cls) -> dict:
        """ gets all the properties

        :returns:
            a deep copy of the  of key/value pairs
        """
        with cls.__lock:
            return cls.__properties.copy()

    @classmethod
    def set(cls, key: Any, value: Any) -> None:
        """ sets a key value pair. The value can be Union(Str, Dict, Tuple, Array)

        :param key: the key string
        :param value: the value of the key
        """
        if key is None or not key or not isinstance(key, str):
            raise ValueError("The key must be a valid str")
        keys = key.split('.')
        _prop_branch = cls.__properties
        _last_key = None
        _last_prop_branch = None
        # from base of the key work up to find where the section doesn't exist
        for _, k in list(enumerate(keys)):
            if k not in _prop_branch:
                break
            _last_prop_branch = _prop_branch
            _last_key = k
            _prop_branch = _prop_branch[k]
        tmp_dict = {}
        # now from the top of the key work back, creating the sections tree
        k = None
        for _, k in reversed(list(enumerate(keys))):
            if isinstance(value, dict):
                tmp_dict = {k: value}
            else:
                tmp_dict[k] = value
            if k is _last_key:
                break
            value = tmp_dict
        if not isinstance(value, dict):
            if isinstance(_last_prop_branch[k], list):
                if isinstance(value, list):
                    _last_prop_branch[k] += value
                else:
                    _last_prop_branch[k].append(value)
            else:
                _last_prop_branch[k] = value
            return
        if _last_prop_branch is None:
            _prop_branch.update(value)
        else:
            cls._add_value(k, value, _last_prop_branch)
        return

    @classmethod
    def remove(cls, key: Any) -> bool:
        """removes a key/value from the in-memory configuration dictionary based on the key

        :param key: the key of the key/value to be removed
            The key should be a dot separated string of keys from root up the tree

        :return:
            True if the key was removed
            False if the key was not found
        """
        del_dict = cls.__properties
        del_path, _, del_key = key.rpartition('.')
        if del_path:
            for part in del_path.split('.'):
                if isinstance(del_dict, dict):
                    del_dict = del_dict.get(part)
                else:
                    return False
        if del_dict is None or del_key not in del_dict:
            return False
        with cls.__lock:
            _ = del_dict.pop(del_key, None)
        return True

    @classmethod
    def load(cls, handler: AbstractPersistHandler, key: str=None, replace: bool=False,
             ignore_key_error: bool=None) -> bool:
        """ loads the properties from the configuration file. allows for multiple configuration
        files to be merged into the properties dictionary, or properties to be refreshed in real time.

        :param handler: the handler to load the configuration from.
        :param key: An optional root key subset of the configuration values (dot format e.g. root.level1).
        :param replace: if the file is to be added to or replaced
        :param ignore_key_error: when providing a key, if the key is not found continue without loading or error
        :return: True if the file was found, opened and loaded, False if an exception was thrown.
        """
        if handler is None or not isinstance(handler, AbstractPersistHandler):
            raise ValueError("The handler must be a concrete implementation of AbstractPersistHandler")
        # if not _path.exists() or not _path.is_file():
        tbl = handler.load_canonical()
        a_data = tbl['properties'].combine_chunks().to_pylist()[0]
        cfg_dict = ast.literal_eval(a_data.decode('utf8'))
        if replace:
            with cls.__lock:
                cls.__properties.clear()
        # Don't copy over the file meta data
        _ = cfg_dict.pop('config_meta', None)
        if isinstance(key, str):
            subset = cfg_dict
            for k in key.split('.'):
                if isinstance(subset, dict) and k in subset:
                    subset = subset.get(k, {})
                else:
                    if isinstance(ignore_key_error, bool) and ignore_key_error:
                        return False
                    raise KeyError(f"The key '{key}' could not be found in the properties canonical when loading.")
            cls.set(key, subset)
        elif replace:
            cls.__properties = cfg_dict
        else:
            for _key, _value in cfg_dict.items():
                # only replace sections that have changed
                cls.set(_key, _value)
        return True

    @classmethod
    def dump(cls, handler: AbstractPersistHandler, key: Any=None):
        """ Dumps the current in-memory configuration to a persistence store.
        Note that this replaces existing content if the file exists. This is particularly
        important is only persisting a single root key.

        The use of the root key option allows for the breakdown of persisted files into
        multiple files for easier management.

        :param handler: the handler to persist the configuration.
        :param key: An optional root key subset of the configuration values.
        """
        if handler is None or not isinstance(handler, AbstractPersistHandler):
            raise ValueError("The handler must be a concrete implementation of AbstractPersistHandler")
        _time = str(datetime.now())
        if not cls.is_key('config_meta.uid') or not cls.is_key('config_meta.create'):
            cls.set('config_meta.uid', str(uuid.uuid4()))
            cls.set('config_meta.create', _time)
        cls.set('config_meta.modify', _time)
        cls.set('config_meta.release', ds_core.__version__)
        # check we want to persist
        if key is None:
            with cls.__lock:
                data = cls.__properties.copy()
        else:
            if not cls.is_key(key):
                raise KeyError("The base key {} does not exist".format(key))
            # from base of the key work up
            keys = key.split('.')
            _pointer = {}
            data = _pointer
            while len(keys) > 0:
                _k = keys.pop(0)
                _pointer[_k] = {}
                _pointer = _pointer[_k]
            _value = cls.get(key)
            if isinstance(_value, str):
                __pointer = _value
            else:
                _pointer.update(cls.get(key))
            data['config_meta'] = cls.get('config_meta')

        # now build the canonical
        enc = str(data).encode('utf8')
        t_data = pa.table([pa.array([enc])], names=['properties'])
        handler.persist_canonical(t_data)
        return

    @staticmethod
    def join(*names, sep: str=None) -> str:
        """Used to create a name string. Can also be used to join paths by passing sep=os.path.sep

        :param names: the names to join
        :param sep: the join separator. Default to '.'
        :return: the names joined with the separator
        """
        _sep = sep if sep is not None else '.'
        return _sep.join(map(str, names))

    @classmethod
    def _remove_all(cls):
        with cls.__lock:
            cls.__properties.clear()

    @classmethod
    def _add_value(cls, key, value, root):
        if key is None:
            return
        for k, v in value.items():
            if isinstance(v, dict) and k in root[key]:
                cls._add_value(k, v, root[key])
            else:
                with cls.__lock:
                    if k in root[key] and isinstance(root[key][k], dict):
                        root[key][k].update(v)
                    else:
                        if k not in root[key] and not isinstance(root[key], dict):
                            root[key] = {}
                        root[key][k] = v
        return



import io
import requests
import os
import pyarrow as pa
import pyarrow.parquet as pq
from pyarrow import csv
from ds_core.components.core_commons import CoreCommons as Commons
from ds_core.handlers.abstract_handlers import AbstractSourceHandler, AbstractPersistHandler
from ds_core.handlers.abstract_handlers import ConnectorContract, HandlerFactory

__author__ = 'Darryl Oatridge'


class BaseSourceHandler(AbstractSourceHandler):
    """ Base PyArrow read only Source Handler. """

    def __init__(self, connector_contract: ConnectorContract):
        """ initialise the Handler passing the connector_contract dictionary """
        super().__init__(connector_contract)
        self._file_state = 0
        self._changed_flag = True

    def supported_types(self) -> list:
        """ The source types supported with this module"""
        return ['parquet', 'csv']

    def load_canonical(self, **kwargs) -> pa.Table:
        """ returns the canonical dataset based on the connector contract. """
        if not isinstance(self.connector_contract, ConnectorContract):
            raise ValueError("The Connector Contract was not been set at initialisation or is corrupted")
        _cc = self.connector_contract
        load_params = kwargs
        load_params.update(_cc.kwargs)  # Update with any kwargs in the Connector Contract
        if load_params.pop('use_full_uri', False):
            file_type = load_params.pop('file_type', 'csv')
            address = _cc.uri
        else:
            load_params.update(_cc.query)  # Update kwargs with those in the uri query
            _, _, _ext = _cc.address.rpartition('.')
            address = _cc.address
            file_type = load_params.pop('file_type', _ext if len(_ext) > 0 else 'csv')
        self.reset_changed()
        # parquet
        if file_type.lower() in ['parquet', 'pqt', 'pq']:
            if _cc.schema.startswith('http'):
                address = io.BytesIO(requests.get(address).content)
            return pq.read_table(address, **load_params)
        # csv
        if file_type.lower() in ['csv', 'gz', 'bz2']:
            _kwargs = {**_cc.query, **_cc.kwargs, **load_params}
            parse_options = _kwargs.get('parse_options', {}).get('parse_options', {})
            parse_options = self.parse_options(**parse_options)
            read_options = _kwargs.get('read_options', {}).get('read_options', {})
            read_options = self.read_options(**read_options)
            if _cc.schema.startswith('http'):
                address = io.BytesIO(requests.get(address).content)
            return csv.read_csv(address, parse_options=parse_options, read_options=read_options)
        raise LookupError('The source format {} is not currently supported'.format(file_type))

    def exists(self) -> bool:
        """ Returns True is the file exists """
        if not isinstance(self.connector_contract, ConnectorContract):
            raise ValueError("The Connector Contract has not been set")
        _cc = self.connector_contract
        if _cc.schema.startswith('http'):
            r = requests.get(_cc.address)
            if r.status_code == 200:
                return True
        if os.path.exists(_cc.address):
            return True
        return False

    def has_changed(self) -> bool:
        """ returns the status of the change_flag indicating if the file has changed since last load or reset"""
        if not self.exists():
            return False
        # maintain the change flag
        _cc = self.connector_contract
        if _cc.schema.startswith('http') or _cc.schema.startswith('git'):
            if not isinstance(self.connector_contract, ConnectorContract):
                raise ValueError("The Pandas Connector Contract has not been set")
            module_name = 'requests'
            _address = _cc.address.replace("git://", "https://")
            if HandlerFactory.check_module(module_name=module_name):
                module = HandlerFactory.get_module(module_name=module_name)
                state = module.head(_address).headers.get('last-modified', 0)
            else:
                raise ModuleNotFoundError(f"The required module {module_name} has not been installed. Please pip "
                                          f"install the appropriate package in order to complete this action")
        else:
            state = os.stat(_cc.address).st_mtime_ns
        if state != self._file_state:
            self._changed_flag = True
            self._file_state = state
        return self._changed_flag

    def reset_changed(self, changed: bool = False):
        """ manual reset to say the file has been seen. This is automatically called if the file is loaded"""
        changed = changed if isinstance(changed, bool) else False
        self._changed_flag = changed

    @staticmethod
    def read_options(**kwargs) -> csv.ReadOptions:
        if kwargs is None or not kwargs:
            return None
        return csv.ReadOptions(**kwargs)

    @staticmethod
    def parse_options(**kwargs) -> csv.ParseOptions:
        if kwargs is None or not kwargs:
            return None
        return csv.ParseOptions(**kwargs)


class BasePersistHandler(BaseSourceHandler, AbstractPersistHandler):
    """ Base PyArrow read/write Persist Handler. """

    def persist_canonical(self, canonical: pa.Table, **kwargs) -> bool:
        """ persists the canonical dataset

        Extra Parameters in the ConnectorContract kwargs:
            - file_type: (optional) the type of the source file. if not set, inferred from the file extension
        """
        if not isinstance(self.connector_contract, ConnectorContract):
            return False
        _uri = self.connector_contract.uri
        return self.backup_canonical(uri=_uri, canonical=canonical, **kwargs)

    def backup_canonical(self, canonical: pa.Table, uri: str, **kwargs) -> bool:
        """ creates a backup of the canonical to an alternative URI

        Extra Parameters in the ConnectorContract kwargs:
            - file_type: (optional) the type of the source file. if not set, inferred from the file extension
            - write_params (optional) a dictionary of additional write parameters directly passed to 'write_' methods
        """
        if not isinstance(self.connector_contract, ConnectorContract):
            return False
        _cc = self.connector_contract
        _address = _cc.parse_address(uri=uri)
        persist_params = kwargs if isinstance(kwargs, dict) else _cc.kwargs
        persist_params.update(_cc.parse_query(uri=uri))
        _, _, _ext = _address.rpartition('.')
        if not self.connector_contract.schema.startswith('http'):
            _path, _ = os.path.split(_address)
            if len(_path) > 0 and not os.path.exists(_path):
                os.makedirs(_path)
        file_type = persist_params.pop('file_type', _ext if len(_ext) > 0 else 'parquet')
        write_params = persist_params.pop('write_params', {})
        # parquet
        if file_type.lower() in ['pq', 'pqt', 'parquet']:
            pq.write_table(canonical, _address, **write_params)
            return True
        # csv
        if file_type.lower() in ['csv', 'gz', 'bz2']:
            for n in canonical.column_names:
                c = canonical.column(n).combine_chunks()
                if pa.types.is_dictionary(c.type):
                    dc = c.dictionary_decode()
                    canonical = Commons.table_append(canonical, pa.table([dc], names=[n]))
                if pa.types.is_nested(c.type):
                    sc = pa.Array.from_pandas(pa.Array.to_pandas(c).astype(str))
                    canonical = Commons.table_append(canonical, pa.table([sc], names=[n]))
            csv.write_csv(canonical, _address, **write_params)
            return True
        # not found
        raise LookupError('The file format {} is not currently supported for write'.format(file_type))

    def remove_canonical(self) -> bool:
        if not isinstance(self.connector_contract, ConnectorContract):
            return False
        _cc = self.connector_contract
        if self.connector_contract.schema.startswith('http'):
            raise NotImplemented("Remove Canonical does not support {} schema based URIs".format(_cc.schema))
        if os.path.exists(_cc.address):
            os.remove(_cc.address)
            return True
        return False



import ast
import unittest
import os
import shutil
from pprint import pprint

import pyarrow as pa
import pyarrow.compute as pc
import pyarrow.parquet as pq
from ds_core.components.core_commons import DataAnalytics, CoreCommons
from ds_core.properties.property_manager import PropertyManager


class CommonsTest(unittest.TestCase):

    def setUp(self):
        os.environ['HADRON_PM_PATH'] = "event://pm_story"
        os.environ['HADRON_DEFAULT_PATH'] = "event://data"
        try:
            os.makedirs(os.environ['HADRON_PM_PATH'])
            os.makedirs(os.environ['HADRON_DEFAULT_PATH'])
        except:
            pass
        PropertyManager._remove_all()

    def tearDown(self):
        try:
            shutil.rmtree('work')
        except:
            pass

    def test_table_append(self):
        t1 = pa.Table.from_pydict({'A': [1,2], 'B': [1,3], 'C': [2,4]})
        t2 = pa.Table.from_pydict({'X': [4,5], 'A': [4,7]})
        result = CoreCommons.table_append(t1, t2)
        self.assertEqual((2, 3),t1.shape)
        self.assertEqual(['A', 'B', 'C'],t1.column_names)
        self.assertEqual((2, 2),t2.shape)
        self.assertEqual(['X', 'A'],t2.column_names)
        self.assertEqual((2, 4),result.shape)
        self.assertEqual(['B', 'C', 'X', 'A'], result.column_names)
        result = CoreCommons.table_append(None, t2)
        self.assertEqual((2, 2),result.shape)
        self.assertEqual(['X', 'A'],result.column_names)
        t3 = pa.Table.from_pydict({'X': [4, 5, 6], 'A': [6, 7, 8]})
        with self.assertRaises(ValueError) as context:
            result = CoreCommons.table_append(t1, t3)
        self.assertTrue("The tables passed are not of equal row size. The first has '2' rows and the second has '3' rows" in str(context.exception))
        with self.assertRaises(ValueError) as context:
            result = CoreCommons.table_append(t1, None)
        self.assertTrue("As a minimum, the second value passed must be a PyArrow Table" in str(context.exception))

    def test_table_flatten_file(self):
        # Individual
        with open('../_data/Individual.txt') as f:
            document = ast.literal_eval(f.read())
        print("\n Individual")
        pprint(document[:1])
        tbl = pa.Table.from_pylist(document)
        f_tbl = CoreCommons.table_flatten(tbl)
        result = CoreCommons.table_nest(f_tbl)
        pprint(result[:1])

        # PartyInteraction
        with open('../_data/PartyInteraction.txt') as f:
            document = ast.literal_eval(f.read())
        print("\n PartyInteraction")
        pprint(document[:1])
        tbl = pa.Table.from_pylist(document)
        f_tbl = CoreCommons.table_flatten(tbl)
        result = CoreCommons.table_nest(f_tbl)
        pprint(result[:1])

        # Organisation
        with open('../_data/Organization.txt') as f:
            document = ast.literal_eval(f.read())
        print("\n Organisation")
        pprint(document[:1])
        tbl = pa.Table.from_pylist(document)
        f_tbl = CoreCommons.table_flatten(tbl)
        result = CoreCommons.table_nest(f_tbl)
        pprint(result[:1])

    def test_table_flatten(self):
        document = [
            {"_id": "I35138", "contactMedium": [{"medium": {"number": "50070028", "type": "mobile"}, "preferred": True}, {"medium": {"emailAddress": "mail@stc.com.kw", "type": "emailAddress"}, "preferred": True}], "gender": "M", "familyName": "Fouad", "givenName": "Fouad", "middleName": "Fouad"},
            {"_id": "I35145", "contactMedium": [{"medium": {"emailAddress": "panneer.rajadurai.c@solutions.com.kw", "type": "EmailAddress"}, "preferred": True}, {"medium": {"number": "51658317", "type": "mobile"}, "preferred": True}, {"medium": {"number": "51658317", "type": "whatsapp"}, "preferred": False}, {"medium": {"number": "51658317", "type": "telegram"}, "preferred": False}, {"medium": {"type": "telephone"}, "role": "AlternateNumber"}], "gender": "M", "familyName": "Jay", "givenName": "Bhuvana", "middleName": ""},
            {"_id": "I35146", "contactMedium": [{"medium": {"emailAddress": "bhuvana.stc21@gmail.com", "type": "EmailAddress"}, "preferred": True}, {"medium": {"type": "mobile"}, "preferred": False}, {"medium": {"type": "whatsapp"}, "preferred": False}, {"medium": {"type": "telegram"}, "preferred": False}], "gender": "F", "familyName": "CORP", "givenName": "TECNOTREE", "middleName": "LTD"},
            {"_id": "I35178", "contactMedium": [{"medium": {"emailAddress": "m.m.alkhoduri@outlook.com", "type": "emailAddress"}, "preferred": True}, {"medium": {"number": "55850055", "type": "mobile"}, "preferred": True}, {"medium": {"number": "55850055", "type": "whatsapp"}, "preferred": False}, {"medium": {"number": "55850055", "type": "telegram"}, "preferred": False}], "gender": "M", "familyName": "", "givenName": "MohammadalKoduri", "middleName": ""},
            {"_id": "I35179", "contactMedium": [{"medium": {"emailAddress": "ahb@bremenintl.com", "type": "emailAddress"}, "preferred": True}, {"medium": {"number": "51500014", "type": "mobile"}, "preferred": True}, {"medium": {"number": "51500014", "type": "whatsapp"}, "preferred": False}, {"medium": {"number": "51500014", "type": "telegram"}, "preferred": False}], "gender": "M", "familyName": "", "givenName": "AhmedBakhiet", "middleName": ""},
            {"_id": "I35180", "contactMedium": [{"medium": {"emailAddress": "test@gmail.com", "type": "emailAddress"}, "preferred": True}], "gender": "M", "familyName": "Admin", "givenName": "FakhrTest", "middleName": ""},
            {"_id": "I35181", "contactMedium": [], "gender": "M", "familyName": "test", "givenName": "test", "nationality": "", "middleName": ""}
        ]
        tbl = pa.Table.from_pylist(document)
        result = CoreCommons.table_flatten(tbl)
        control = CoreCommons.filter_headers(result, regex='nest_list_', drop=True)
        self.assertCountEqual(['middleName', 'gender', '_id', 'givenName', 'familyName'], control)
        control = len(CoreCommons.filter_headers(result, regex='nest_list_'))
        self.assertEqual(16, control)

    def test_table_nest(self):
        document = [
            {"_id": "I35138", "contactMedium": [{"medium": {"number": "50070028", "type": "mobile"}, "preferred": True}, {"medium": {"emailAddress": "mail@stc.com.kw", "type": "emailAddress"}, "preferred": True}], "gender": "M", "familyName": "Fouad", "givenName": "Fouad", "middleName": "Fouad"},
            {"_id": "I35145", "contactMedium": [{"medium": {"emailAddress": "panneer.rajadurai.c@solutions.com.kw", "type": "EmailAddress"}, "preferred": True}, {"medium": {"number": "51658317", "type": "mobile"}, "preferred": True}, {"medium": {"number": "51658317", "type": "whatsapp"}, "preferred": False}, {"medium": {"number": "51658317", "type": "telegram"}, "preferred": False}, {"medium": {"type": "telephone"}, "role": "AlternateNumber"}], "gender": "M", "familyName": "Jay", "givenName": "Bhuvana", "middleName": ""},
            {"_id": "I35146", "contactMedium": [{"medium": {"emailAddress": "bhuvana.stc21@gmail.com", "type": "EmailAddress"}, "preferred": True}, {"medium": {"type": "mobile"}, "preferred": False}, {"medium": {"type": "whatsapp"}, "preferred": False}, {"medium": {"type": "telegram"}, "preferred": False}], "gender": "F", "familyName": "CORP", "givenName": "TECNOTREE", "middleName": "LTD"},
            {"_id": "I35178", "contactMedium": [{"medium": {"emailAddress": "m.m.alkhoduri@outlook.com", "type": "emailAddress"}, "preferred": True}, {"medium": {"number": "55850055", "type": "mobile"}, "preferred": True}, {"medium": {"number": "55850055", "type": "whatsapp"}, "preferred": False}, {"medium": {"number": "55850055", "type": "telegram"}, "preferred": False}], "gender": "M", "familyName": "", "givenName": "MohammadalKoduri", "middleName": ""},
            {"_id": "I35179", "contactMedium": [{"medium": {"emailAddress": "ahb@bremenintl.com", "type": "emailAddress"}, "preferred": True}, {"medium": {"number": "51500014", "type": "mobile"}, "preferred": True}, {"medium": {"number": "51500014", "type": "whatsapp"}, "preferred": False}, {"medium": {"number": "51500014", "type": "telegram"}, "preferred": False}], "gender": "M", "familyName": "", "givenName": "AhmedBakhiet", "middleName": ""},
            {"_id": "I35180", "contactMedium": [{"medium": {"emailAddress": "test@gmail.com", "type": "emailAddress"}, "preferred": True}], "gender": "M", "familyName": "Admin", "givenName": "FakhrTest", "middleName": ""},
            {"_id": "I35181", "contactMedium": [], "gender": "M", "familyName": "test", "givenName": "test", "nationality": "", "middleName": ""}
        ]
        tbl = CoreCommons.table_flatten(pa.Table.from_pylist(document))
        result = CoreCommons.table_nest(tbl)
        self.assertCountEqual(['_id', 'gender', 'familyName', 'givenName', 'middleName', 'contactMedium'], result[0].keys())
        self.assertIsInstance(result[0].get('contactMedium'), list)
        self.assertEqual(len(result[0].get('contactMedium')), 2)
        self.assertEqual(len(result[1].get('contactMedium')), 5)
        self.assertEqual(len(result[2].get('contactMedium')), 4)
        self.assertCountEqual(['_id', 'gender', 'familyName', 'givenName', 'middleName'], result[6].keys())
        pprint(result[:1])

    def test_table_nest_list(self):
        document = [{'a' : [{'b': 1, 'c': [{'z': 6}]},
                            {'c': [{'z': 4}]}]
                     }]
        tbl = CoreCommons.table_flatten(pa.Table.from_pylist(document))
        self.assertEqual(tbl.column_names , ['a.nest_list_0.b', 'a.nest_list_0.c.nest_list_0.z', 'a.nest_list_1.c.nest_list_0.z'] )
        result = CoreCommons.table_nest(tbl)
        self.assertEqual(result,  [{'a': [{'b': 1, 'c': [{'z': 6}]}, {'c': [{'z': 4}]}]}])

    def test_nest_list_order(self):
        document = [{'name': [{'state': 'AZ'},]},
                    {'name': [{'city': 'Fredville'},
                             {'state': 'VA'},]},
                    {'name': [{'state': 'TX'},
                              {'city': 'Georgetown'},]},]
        tbl = CoreCommons.table_flatten(pa.Table.from_pylist(document), flatten_list=False)
        print(CoreCommons.tbl)




    def test_table_nest_flatten(self):
        num = pa.array([1.0, 12.0, 5.0, None], pa.float64())
        val = pa.array([7, None, 3, 5], pa.int64())
        date = pc.strptime(["2023-01-02 04:49:06", "2023-01-02 04:57:12", None, "2023-01-02 05:23:50"], format='%Y-%m-%d %H:%M:%S', unit='ns')
        text = pa.array(["Blue", "Green", None, 'Red'], pa.string())
        binary = pa.array([True, True, None, False], pa.bool_())
        cat = pa.array([None, 'M', 'F', 'M'], pa.string()).dictionary_encode()
        tbl = pa.table([num, val, date, text, binary, cat],
                       names=['num', 'int', 'date', 'text', 'bool', 'cat'])
        result = CoreCommons.table_nest(tbl)
        pprint(result)
        print('')
        t = pa.Table.from_pylist(result)
        print(t.schema)
        print('')
        result = CoreCommons.table_flatten(t)
        print(result.schema)


    def test_column_precision(self):
        num = pa.array([1.471, 12, 5.33, None], pa.float64())
        text = pa.array(["Blue", "Green", None, 'Blue'], pa.string())
        val = pa.array([1, 0, 1, None], pa.int64())
        result = CoreCommons.column_precision(num)
        self.assertEqual(3, result)
        result = CoreCommons.column_precision(val)
        self.assertEqual(0, result)
        with self.assertRaises(ValueError) as context:
            result = CoreCommons.column_precision(text)
        self.assertTrue("The array should be numeric, type 'string' sent." in str(context.exception))

    def test_column_join(self):
        a = pa.array([1,2,3,4,5])
        b = pa.array(list('abcde'))
        c = CoreCommons.column_join(a,b)
        self.assertEqual(['1a', '2b', '3c', '4d', '5e'], c.to_pylist())
        c = CoreCommons.column_join(a,b,sep='_')
        self.assertEqual(['1_a', '2_b', '3_c', '4_d', '5_e'], c.to_pylist())

    def test_array_cast(self):
        num = pa.array([1.0, 12.0, 5.0, None], pa.float64())
        date = pc.strptime(["2023-01-02 04:49:06", "2023-01-02 04:57:12", None, "2023-01-02 05:23:50"], format='%Y-%m-%d %H:%M:%S', unit='ns')
        text = pa.array(["Blue", "Green", None, 'Blue'], pa.string())
        bool1 = pa.array([1, 0, 1, None], pa.int64())
        bool2 = pa.array(['true', 'true', None, 'false'], pa.string())
        bool3 = pa.array([None, 'no', 'yes', 'yes'], pa.string())
        # ensure it doesn't fall over
        for c in [num, date, text, bool1, bool2, bool3]:
            for ty in [pa.string(),pa.int64(),pa.float64(),pa.timestamp('ns'),pa.bool_(),pa.null(),]:
                _ = CoreCommons.column_cast(c, ty)
        self.assertEqual(pa.int64(), CoreCommons.column_cast(num, pa.int64()).type)
        self.assertEqual(pa.timestamp('ns'), CoreCommons.column_cast(date, pa.timestamp('ns')).type)
        self.assertEqual(pa.string(), CoreCommons.column_cast(text, pa.string()).type)
        self.assertEqual(pa.bool_(), CoreCommons.column_cast(bool1, pa.bool_()).type)
        self.assertEqual(pa.bool_(), CoreCommons.column_cast(bool2, pa.bool_()).type)
        self.assertEqual(pa.string(), CoreCommons.column_cast(bool3, pa.bool_()).type)

    def test_table_cast_us_date(self):
        str_dt = pa.array(["01-16-2023", "03-07-2023", None, "11-05-2023"], pa.string())
        tbl = pa.table([str_dt], names=['str_dt'])
        result = CoreCommons.table_cast(tbl, dt_format="%m-%d-%Y").combine_chunks()
        control = pa.schema([('str_dt', pa.timestamp('ns'))])
        self.assertEqual(control, result.schema)

    def test_table_cast(self):
        num = pa.array([1.0, 12.0, 5.0, None], pa.float64())
        date = pc.strptime(["2023-01-02 04:49:06", "2023-01-02 04:57:12", None, "2023-01-02 05:23:50"], format='%Y-%m-%d %H:%M:%S', unit='ns')
        str_dt = pa.array(["2023-01-16", "2023-03-07", None, "2023-11-05"], pa.string())
        value = pa.array([None, '1.5', '3.2', '2.0'], pa.string())
        text = pa.array(["Blue", "Green", None, 'Red'], pa.string())
        bool1 = pa.array([1, 0, 1, None], pa.int64())
        bool2 = pa.array(['true', 'true', None, 'false'], pa.string())
        bool3 = pa.array([None, 'M', 'F', 'M'], pa.string())
        tbl = pa.table([num, value, date, str_dt, text, bool1, bool2, bool3],
                       names=['num', 'value', 'date', 'str_dt', 'text', 'bool1', 'bool2', 'bool3'])
        result = CoreCommons.table_cast(tbl).combine_chunks()
        control = pa.schema([('num', pa.int64()),
                             ('value', pa.float64()),
                             ('date', pa.timestamp('ns')),
                             ('str_dt', pa.timestamp('ns')),
                             ('text', pa.dictionary(pa.int32(), pa.string())),
                             ('bool1', pa.bool_()),
                             ('bool2', pa.bool_()),
                             ('bool3', pa.dictionary(pa.int32(), pa.string())),
                             ])
        self.assertEqual(control, result.schema)
        result = CoreCommons.table_cast(result, inc_cat=False).combine_chunks()
        control = pa.schema([('num', pa.int64()),
                             ('value', pa.float64()),
                             ('date', pa.timestamp('ns')),
                             ('str_dt', pa.timestamp('ns')),
                             ('text', pa.string()),
                             ('bool1', pa.bool_()),
                             ('bool2', pa.bool_()),
                             ('bool3', pa.string()),
                             ])
        self.assertEqual(control, result.schema)
        result = CoreCommons.table_cast(tbl, cat_max=2).combine_chunks()
        control = pa.schema([('num', pa.int64()),
                             ('value', pa.float64()),
                             ('date', pa.timestamp('ns')),
                             ('str_dt', pa.timestamp('ns')),
                             ('text', pa.string()),
                             ('bool1', pa.bool_()),
                             ('bool2', pa.bool_()),
                             ('bool3', pa.dictionary(pa.int32(), pa.string())),
                             ])
        self.assertEqual(control, result.schema)
        result = CoreCommons.table_cast(tbl, cat_max=2, inc_cat=False, inc_bool=False).combine_chunks()
        control = pa.schema([('num', pa.int64()),
                             ('value', pa.float64()),
                             ('date', pa.timestamp('ns')),
                             ('str_dt', pa.timestamp('ns')),
                             ('text', pa.string()),
                             ('bool1', pa.int64()),
                             ('bool2', pa.string()),
                             ('bool3', pa.string()),
                             ])
        self.assertEqual(control, result.schema)

    def test_filter_headers(self):
        tbl = pq.read_table("../_data/sample_types.parquet")
        result = CoreCommons.filter_headers(tbl)
        self.assertEqual(tbl.column_names, result)
        result = CoreCommons.filter_headers(tbl, headers=['num', 'string', 'cat'])
        self.assertCountEqual(['num', 'string', 'cat'], result)
        result = CoreCommons.filter_headers(tbl, regex=['num', 'int'])
        self.assertCountEqual(['num','int','num_null','int_null','nulls_int'], result)
        result = CoreCommons.filter_headers(tbl, regex=['null', 'num', 'int', 'date', 'nest'], drop=True)
        self.assertCountEqual(['binary', 'cat', 'bool', 'string'], result)
        result = CoreCommons.filter_headers(tbl, headers=['num', 'int', 'string'], regex='i')
        self.assertCountEqual(['int', 'string'], result)
        result = CoreCommons.filter_headers(tbl, d_types=['is_boolean', 'is_string'])
        self.assertCountEqual(['bool', 'string', 'bool_null', 'string_null', 'nulls_str'], result)
        result = CoreCommons.filter_headers(tbl, d_types=[pa.bool_(), pa.string()])
        self.assertCountEqual(['bool', 'string', 'bool_null', 'string_null', 'nulls_str'], result)
        result = CoreCommons.filter_headers(tbl, headers=['num','int','string'], d_types=['is_boolean', 'is_string'])
        self.assertCountEqual(['string'], result)

    def test_filter_columns(self):
        tbl = pq.read_table("../_data/sample_types.parquet")
        result = CoreCommons.filter_columns(tbl, headers=['num', 'date', 'string'], regex='t', d_types=[pa.string()])
        self.assertCountEqual(['string'], result.column_names)
        result = CoreCommons.filter_columns(tbl, headers='num')
        self.assertCountEqual(['num'], result.column_names)

    def test_list_formatter(self):
        sample = {'A': [1,2], 'B': [1,2], 'C': [1,2]}
        result = CoreCommons.list_formatter(sample)
        self.assertEqual(list("ABC"), result)
        result = CoreCommons.list_formatter(sample.keys())
        self.assertEqual(list("ABC"), result)
        result = CoreCommons.list_formatter("A")
        self.assertEqual(['A'], result)

    def test_presision_scale(self):
        sample = 1
        self.assertEqual((1, 0), CoreCommons.precision_scale(sample))
        sample = 729.0
        self.assertEqual((3, 0), CoreCommons.precision_scale(sample))
        sample = 729.4
        self.assertEqual((4, 1), CoreCommons.precision_scale(sample))
        sample = 0.456
        self.assertEqual((4, 3), CoreCommons.precision_scale(sample))
        sample = 2784.45612367432
        self.assertEqual((14, 10), CoreCommons.precision_scale(sample))
        sample = -3.72
        self.assertEqual((3, 2), CoreCommons.precision_scale(sample))
        sample = -4
        self.assertEqual((1, 0), CoreCommons.precision_scale(sample))

    def test_list_dup(self):
        seq = ['A', 'B', 'B', 'C', 'C', 'D']
        result = CoreCommons.list_dup(seq=seq)
        self.assertCountEqual(['B', 'C'], result)
        seq = ['A', 'B', 'C', 'D']
        result = CoreCommons.list_dup(seq=seq)
        self.assertCountEqual([], result)

    def tests_list_beta_find(self):
        seq = ['A', 'B', 'C', 'D', 'E']
        result = CoreCommons.list_search(seq, 'A')
        self.assertEqual(0, result)
        result = CoreCommons.list_search(seq, 'C')
        self.assertEqual(2, result)
        result = CoreCommons.list_search(seq, 'E')
        self.assertEqual(4, result)

    def test_list_standardize(self):
        seq = [100, 75, 50, 25, 0]
        result = CoreCommons.list_standardize(seq=seq)
        self.assertEqual(0.0, result[2])
        self.assertEqual(0.0, result[1] + result[3])
        self.assertEqual(0.0, result[0] + result[4])

    def test_list_normalize_robust(self):
        seq = [100, 75, 50, 25, 0]
        result = CoreCommons.list_normalize_robust(seq=seq)
        self.assertEqual(result, [1.5, 1.0, 0.5, 0.0, -0.5])
        # outlier
        seq = [1,1,2,100,1,2,1,1,1]
        result = CoreCommons.list_normalize_robust(seq=seq)
        self.assertEqual(result, [0.0, 0.0, 1.0, 99.0, 0.0, 1.0, 0.0, 0.0, 0.0])


    def test_list_normalize(self):
        seq = [100, 75, 50, 25, 0]
        a = 0
        b = 1
        result = CoreCommons.list_normalize(seq=seq, a=a, b=b)
        self.assertEqual([1.0, 0.75, 0.5, 0.25, 0.0], result)
        a = -1
        b = 1
        result = CoreCommons.list_normalize(seq=seq, a=a, b=b)
        self.assertEqual([1.0, 0.5, 0, -0.5, -1], result)

    def test_diff(self):
        a = [1,2,3,4]
        b = [2,3,4,6,7]
        self.assertEqual([1, 6, 7], CoreCommons.list_diff(a, b))
        self.assertEqual([1], CoreCommons.list_diff(a, b, symmetric=False))

    def test_intersect(self):
        a = [1,2,3,3]
        b = [2,3,4,6,7]
        self.assertEqual([2, 3], CoreCommons.list_intersect(a, b))

    def test_is_list_equal(self):
        a = [1, 4, 2, 1, 4]
        b = [4, 2, 4, 1, 1]
        self.assertTrue(CoreCommons.list_equal(a, b))
        b = [4, 2, 4, 2, 1]
        self.assertFalse(CoreCommons.list_equal(a, b))

    def test_resize_list(self):
        seq = [1,2,3,4]
        for size in range(10):
            result = CoreCommons.list_resize(seq=seq, resize=size)
            self.assertEqual(size, len(result))
            if len(result) >= 4:
                self.assertEqual(seq, CoreCommons.list_unique(result))

    def test_dict_builder(self):
        result = CoreCommons.param2dict()
        self.assertEqual({}, result)
        result = CoreCommons.param2dict(a=1, b='B')
        self.assertEqual({'a': 1, 'b': 'B'}, result)
        result = CoreCommons.param2dict(a=1, b=[1, 2, 3])
        self.assertEqual({'a': 1, 'b': [1, 2, 3]}, result)
        result = CoreCommons.param2dict(a={'A': 1})
        self.assertEqual({'a': {'A': 1}}, result)
        result = CoreCommons.param2dict(a=1, b=None)
        self.assertEqual({'a': 1}, result)

    def test_dict_with_missing(self):
        default = 'no value'
        sample = CoreCommons.dict_with_missing({}, default)
        result = sample['key']
        self.assertEqual(default, result)

    def test_bytestohuman(self):
        result = CoreCommons.bytes2human(1024)
        self.assertEqual('1.0KB', result)
        result = CoreCommons.bytes2human(1024 ** 2)
        self.assertEqual('1.0MB', result)
        result = CoreCommons.bytes2human(1024 ** 3)
        self.assertEqual('1.0GB', result)

    def test_validate_date(self):
        str_date = '2017/01/23'
        self.assertTrue(CoreCommons.valid_date(str_date))
        str_date = '2017/23/01'
        self.assertTrue(CoreCommons.valid_date(str_date))
        str_date = '23-01-2017'
        self.assertTrue(CoreCommons.valid_date(str_date))
        str_date = '01-21-2017'
        self.assertTrue(CoreCommons.valid_date(str_date))
        str_date = 'NaT'
        self.assertFalse(CoreCommons.valid_date(str_date))
        str_date = ''
        self.assertFalse(CoreCommons.valid_date(str_date))
        str_date = '01-21-2017 21:12:46'
        self.assertTrue(CoreCommons.valid_date(str_date))
        str_date = '01/21/2017 21:12:46.000000'
        self.assertTrue(CoreCommons.valid_date(str_date))

    def test_analytics(self):
        analysis = {'intent': {'categories': ['a', 'b'], 'dtype': 'category'},
                    'patterns': {'relative_freq': [.6, .4], 'unique_count': 2}}
        result = DataAnalytics(analysis)
        self.assertEqual(['a', 'b'], result.get('intent').get('categories', []))
        self.assertEqual([], result.get('no_name', []))
        self.assertEqual({}, result.get('no_name'))
        self.assertEqual(['a', 'b'], result.intent.categories)
        self.assertEqual('category', result.intent.dtype)
        self.assertEqual(['categories', 'dtype'], result.intent.elements())

    def test_analysis_build(self):
        result = DataAnalytics.build_category(header='gender', lower=0.1, nulls_list=['', ' '])
        self.assertEqual({'gender': {'lower': 0.1, 'nulls_list': ['', ' ']}}, result)
        result = DataAnalytics.build_number(header='age', lower=0.1, upper=100, precision=2)
        self.assertEqual({'age': {'lower': 0.1, 'upper': 100, 'precision': 2}}, result)
        result = DataAnalytics.build_date(header='birth', granularity=4, year_first=True)
        self.assertEqual({'birth': {'granularity': 4, 'year_first': True}}, result)

    def test_label_gen(self):
        gen = CoreCommons.label_gen()
        result = next(gen)
        self.assertTrue('A', result)
        result = [next(gen) for x in range(5)]
        self.assertTrue(['B', 'C', 'D', 'E', 'F'], result)

    def test_raise(self):
        with self.assertRaises(KeyError) as context:
            env = os.environ['NoEnvValueTest']
        self.assertTrue("'NoEnvValueTest'" in str(context.exception))


if __name__ == '__main__':
    unittest.main()
